/* zodiac - advanced dns spoofer
 *
 * by scut / teso
 *
 * dns tool routines
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <libnet.h>
#include "common.h"
#include "dns.h"
#include "dns-build.h"
#include "dns-spoof.h"
#include "dnsq.h"
#include "gui-config.h"
#include "network.h"
#include "output.h"
#include "packet.h"
#include "zodiac.h"


/* dt_bind_version
 *
 * try to retrieve a version number from a dns server with the host name
 * `host' which is running the bind named.
 *
 * this would be easily done using a fixed buffer and an udp socket, but
 * we want to do it with style, oh yeah =) (in other words i didn't write
 * this routines to lever me down to udp sockets again ;)
 *
 * return an allocated string with the server response
 * return an allocated string "unknown" if the version couldn't be retrieved
 * return NULL on failure (no response)
 */

char *
dt_bind_version (char *host)
{
	int			n,	/* temporary return value */
				desc;	/* dns packet filter descriptor */
	char			*my_ip;	/* local ip */
	dns_pdata		*qpack;	/* query packet data */
	struct in_addr		ip_src,
				ip_dst;
	unsigned short int	s_port;	/* source port */
	unsigned short int	dns_id;	/* query id */
	struct timeval		tvs = { 20, 0 };	/* timeout :) */
	dq_packet		*catch;	/* catched packet */
//	ip_hdr			*ip;	/* pointer to ip header */
//	udp_hdr			*udp;	/* pointer to udp header */
//	dns_hdr			*dns;	/* pointer to dns header */
//	unsigned char		*dns_d;	/* pointer to dns data part within packet */

	/* seed random data a bit
	 */
	s_port = (libnet_get_prand (PRu16) % (65534 - 1024)) + 1025;
	dns_id = libnet_get_prand (PRu16);

	/* set ip's
	 */
	my_ip = net_getlocalip ();
	ip_dst.s_addr = net_resolve (host);
	ip_src.s_addr = net_resolve (my_ip);
	free (my_ip);

	/* install filter, then send packet
	 */
	desc = dq_filter_install (
		ip_dst,			/* dns responses from remote server */
		ip_src,			/* ourself */
		53,			/* responses always from 53 */
		s_port,			/* "our" port ;) */
		1, dns_id, dns_id,	/* it's our response, with our id :) */
		NULL);			/* NULL, it's unique anyway =) */

	if (desc == -1) {
		return (NULL);
	}

	/* build query packet
	 */
	qpack = dns_build_new ();
	dns_build_q (qpack, "VERSION.BIND", T_TXT, C_CHAOS);

	/* fire our little putty =)
	 */
	n = dns_packet_send (my_ip, host, s_port, 53, dns_id, 0, 1, 0, 0, 0, qpack, 1);
	dns_build_destroy (qpack);

	/* in case the packet failed, just return
	 */
	if (n != 0) {
		dq_filter_uninstall (desc);
		return (NULL);
	}

	/* catch the response packet, hopefully =)
	 */
	n = dq_filter_wait (desc, &tvs);
	if (n == 0) {
		dq_filter_uninstall (desc);
		return (NULL);
	}

	/* now we got an response, parse it B-)
	 */
	catch = dq_p_get (desc);
	dq_filter_uninstall (desc);
	if (catch == NULL)
		return (NULL);

	hexdump ("response-bind-version", catch->packet, catch->plen);
/*
	pq_offset (catch->packet, &ip, &udp, &dns, &dns_d);
*/
	dq_p_free (catch);

	return (NULL);
}


